/* ARCADEmini assembler (unix)
 * version 1.4 date 2004-03-15
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2004 1stein <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>

#include "id_list.h"
#include "node.h"
#include "parse_.h"
#include "scan.h"
#include "get_val.h"
#include "textout.h"
#include "global.h"

//print current number of errors and warnings
void print_err_warn( )
{
  char text[64] = "";

  //append number of errors to text
  if( error_cnt == 0 )
    strcat( text, "no errors" );
  else if( error_cnt == 1 )
    strcat( text, "1 error" );
  else
    sprintf( text + strlen( text ), "%d errors", error_cnt );

  //append delimiter to text
  strcat( text, ", " );

  //append number of warnings to text
  if( warning_cnt == 0 )
    strcat( text, "no warnings" );
  else if( warning_cnt == 1 )
    strcat( text, "1 warning" );
  else
    sprintf( text + strlen( text ), "%d warnings", warning_cnt );

  //output text in a box
  print_box( text );
}

//enumeration function to check declaration of identifiers
void ident_check_enum_func( st_id_list_entry * p_entry )
{
  int ret_val;

  //identifier is not declared
  if( p_entry->decl_column == 0 || p_entry->decl_line == 0 )
  {
    //print error message
    printf( "ERROR: %s \"%s%s\" used in column %lu of line %lu is not declared\n",
      id_list_type2typestr( p_entry->type ), id_list_type2prefix( p_entry->type ), p_entry->p_name,
      p_entry->firstuse_column, p_entry->firstuse_line );
    error_cnt++;
  }

  //identifier is not used
  if( p_entry->firstuse_column == 0 || p_entry->firstuse_line == 0 )
  {
    //print warning message
    printf( "WARNING: %s \"%s%s\" declared in column %lu of line %lu is not used\n",
      id_list_type2typestr( p_entry->type ), id_list_type2prefix( p_entry->type ), p_entry->p_name,
      p_entry->decl_column, p_entry->decl_line );
    warning_cnt++;
  }

  //identifier is declared
  if( p_entry->decl_column != 0 && p_entry->decl_line != 0 )
  {
    //try to get value of identifier
    ret_val = get_val_ident( p_entry );
    if( ret_val > 0 ) //cyclical dependency
    {
      //print error message
      printf( "ERROR: %s \"%s%s\" declared in column %lu of line %lu depends cyclically on itself\n",
        id_list_type2typestr( p_entry->type ), id_list_type2prefix( p_entry->type ), p_entry->p_name,
        p_entry->decl_column, p_entry->decl_line );
      error_cnt++;
    }
    if( ret_val < 0 ) //can't get value
    {
      //print error message
      printf( "ERROR: can't get value for %s \"%s%s\" declared in column %lu of line %lu\n",
        id_list_type2typestr( p_entry->type ), id_list_type2prefix( p_entry->type ), p_entry->p_name,
        p_entry->decl_column, p_entry->decl_line );
      error_cnt++;
    }
  } //if( p_entry->decl_column ...
}

//enumeration function to write identifiers
FILE * ident_write_f; //open file to write identifiers to (a "parameter" of this function)
void ident_write_enum_func( st_id_list_entry * p_entry )
{
  unsigned long val;
  unsigned char reg_no;
  unsigned long addr_sec;
  unsigned short addr_ofs;

  //output identifier and its value
  switch( p_entry->type )
  {
    case id_list_con:
      if( get_val_read_ident_con( p_entry, &val ) == 0 )
        fprintf( ident_write_f, "%s\t0x%08lX\n", p_entry->p_name, val );
      else
      {
        printf( "ERROR: can't get value of constant identifier \"%s\"\n", p_entry->p_name );
        error_cnt++;
      }
      break;
    case id_list_reg:
      if( get_val_read_ident_reg( p_entry, &reg_no ) == 0 )
        fprintf( ident_write_f, "$%s\t0x%02X\n", p_entry->p_name, reg_no );
      else
      {
        printf( "ERROR: can't get value of register identifier \"$%s\"\n", p_entry->p_name );
        error_cnt++;
      }
      break;
    case id_list_addr:
      if( get_val_read_ident_addr( p_entry, &addr_sec, &addr_ofs ) == 0 )
        fprintf( ident_write_f, "@%s\t%08lX:%03X\n", p_entry->p_name, addr_sec, addr_ofs );
      else
      {
        printf( "ERROR: can't get value of address identifier \"@%s\"\n", p_entry->p_name );
        error_cnt++;
      }
      break;
  }
}

//assemble
// - get values of command parameters
// - assemble commands
void assemble( )
{
  st_node * p_node, * p_cmd, * p_param;
  unsigned char param_no;
  int ret_val, cmd_params_ok;
  unsigned int cmd_bit, bits_written;

  //loop though all commands in list
  for( p_node = p_list; p_node != NULL; p_node = p_node->data.list.p_list )
  {
    //check if still a list node
    if( p_node->type != node_list )
    {
      printf( "INTERNAL ERROR: list contains a non-list node\n" );
      error_cnt++;
      return;
    }

    //get command node
    p_cmd = p_node->data.list.p_entry;
    while( p_cmd != NULL && p_cmd->type == node_lab )
      p_cmd = p_cmd->data.lab.p_real_node;

    //command node found 
    //(could be also a constant or a register node - we do not want them here)
    if( p_cmd != NULL && p_cmd->type == node_cmd )
    {

      //command data no yet available
      //(e.g. already availabale for string "DB" commmand)
      if( !p_cmd->data.cmd.data_ok )
      {

        //get values of all parameters of this command
        cmd_params_ok = 1;
        for( param_no = 0; param_no < p_cmd->data.cmd.param_cnt; param_no++ )
        {
          //get parameter node
          p_param = p_cmd->data.cmd.params[param_no].p_node;
          //parameter available
          if( p_param != NULL )
          {
            //try to get value of this parameter
            ret_val = get_val_node( p_param );
            if( ret_val > 0 ) //cyclical dependency
            {
              cmd_params_ok = 0;
              //print error message
              printf( "ERROR: parameter %d of command in column %lu of line %lu depends cyclically on itself\n",
                param_no, p_param->create_column, p_param->create_line );
              error_cnt++;
            }
            if( ret_val < 0 ) //can't get value
            {
              cmd_params_ok = 0;
              //print error message
              printf( "ERROR: can't get value for parameter %d of command in column %lu of line %lu\n",
                param_no, p_param->create_column, p_param->create_line );
              error_cnt++;
            }
          }
        } //for( param_no ...

        //command parameters are okay
        if( cmd_params_ok )
        {

          //assemble command
          cmd_bit = 0;
          for( param_no = 0; param_no < p_cmd->data.cmd.param_cnt; param_no++ )
          {
            //read parameter into command buffer
            if( get_val_get_cmd_param( p_cmd, param_no, p_cmd->data.cmd.data, p_cmd->data.cmd.len, cmd_bit, &bits_written ) != 0 )
            {
              printf( "INTERNAL ERROR: can't assemble command in column %lu of line %lu (parameter %d)\n",
                p_cmd->create_column, p_cmd->create_line, param_no );
              error_cnt++;
              break;
            }
            cmd_bit += bits_written; //move behind written data
          }
          //command successfully assembled
          if( param_no == p_cmd->data.cmd.param_cnt )
          {
            //length of data does not match length of command
            if( cmd_bit != (unsigned int)(p_cmd->data.cmd.len << 3) )
            {
              printf( "INTERNAL ERROR: length of parameters does not match expected length of command in column %lu of line %lu\n",
                p_cmd->create_column, p_cmd->create_line );
              error_cnt++;
            }
            //length of data matches length of command
            else
              //remember that command data is complete now
              p_cmd->data.cmd.data_ok = 1;
          }        
        
        } //if( cmd_params_ok )
      } //if( !p_cmd->data.cmd.data_ok )

    } //if( p_cmd != NULL ...
  } //for( p_node ...
}

//help routines for flash output
static int flash_out_byte( unsigned char data, FILE * flashout_f )
{
  if( fputc( ((data >> 4) & 0x0F) + 'A', flashout_f ) == EOF )
    return -1;
  if( fputc( (data & 0x0F) + 'A', flashout_f ) == EOF )
    return -1;
  return 0;
}
static unsigned int flash_out_data( unsigned char * p_data, unsigned int data_len, FILE * flashout_f )
{
  unsigned int i;
  for( i = 0; i < data_len; i++ )
    if( flash_out_byte( p_data[i], flashout_f ) != 0 )
      break;
  return i;
}

//write output
void output( FILE * textout_f, FILE * flashout_f, FILE * output_f )
{
  char identification[0x20] = "ARCADEmini 26x20/16 ambin 1.0   ";
  st_node * p_node, * p_cmd;
  unsigned int i;
  unsigned __int64 addr, pos;
  unsigned long sectors;


  //flash output: header
  if( flashout_f != NULL )
    fwrite( "amdata=", 1, 7, flashout_f );


  //write identification to first sector
  if( fwrite( identification, 1, sizeof( identification ), output_f ) != sizeof( identification ) )
  {
    printf( "ERROR: cannot write to output file\n" );
    error_cnt++;
    return;
  }
  if( flashout_f != NULL )
  {
    if( flash_out_data( identification, sizeof( identification ), flashout_f ) != sizeof( identification ) )
    {
      printf( "ERROR: cannot write to output file\n" );
      error_cnt++;
      return;
    }
  }   
  //fill rest of sector with 0xFF
  for( pos = 0x20; pos < 0x200; pos++ )
  {
    if( fputc( 0xFF, output_f ) == EOF )
      break;
    if( flashout_f != NULL )
      if( flash_out_byte( 0xFF, flashout_f ) != 0 )
        break;
  }
  if( pos < 0x200 )
  {
    printf( "ERROR: cannot write to output file\n" );
    error_cnt++;
    return;
  }


  //loop though all commands in list
  addr = 0;
  pos = 0;
  for( p_node = p_list; p_node != NULL; p_node = p_node->data.list.p_list )
  {
    //check if still a list node
    if( p_node->type != node_list )
    {
      printf( "INTERNAL ERROR: list contains a non-list node\n" );
      error_cnt++;
      return;
    }

    //get command node
    p_cmd = p_node->data.list.p_entry;
    while( p_cmd != NULL && p_cmd->type == node_lab )
      p_cmd = p_cmd->data.lab.p_real_node;

    //command node found and length is not 0
    //(could be also a constant or a register node - we do not want them here)
    if( p_cmd != NULL && p_cmd->type == node_cmd && p_cmd->data.cmd.len > 0 )
    {
      //command was assembled successfully
      if( p_cmd->data.cmd.data_ok )
      {
        //output command as text
        if( textout_f != NULL )
        {
          fprintf( textout_f, "%08lX:%03X", p_cmd->data.cmd.addr_sec, p_cmd->data.cmd.addr_ofs );
          for( i = 0; i < p_cmd->data.cmd.len; i++ )
            fprintf( textout_f," %02X", p_cmd->data.cmd.data[i] );
          fputc( '\n', textout_f );
          //output as human readable command if no DB/DW/DD-command
          if( p_cmd->data.cmd.param_cnt > 0 &&p_cmd->data.cmd.params[0].p_node == NULL )
            textout_textout( textout_f, p_cmd->data.cmd.data, p_cmd->data.cmd.len );
        }
        //output command in binary
        addr = (unsigned __int64)p_cmd->data.cmd.addr_sec << 9 | (unsigned __int64)p_cmd->data.cmd.addr_ofs;
        if( addr < pos ) //address decreased - this is theoretically impossible
        {
          printf( "INTERNAL ERROR: address decreased during output of command in column %lu of line %lu is missing\n", p_cmd->create_column, p_cmd->create_line );
          error_cnt++;
          return;
        }
        else
        {
          //fill free space to next command with 0x00
          for( ; pos < addr; pos++ )
          {
            if( fputc( 0x00, output_f ) == EOF )
              break;
            if( flashout_f != NULL )
              if( flash_out_byte( 0x00, flashout_f ) != 0 )
                break;
          }
          if( pos < addr )
          {
            printf( "ERROR: cannot write to output file\n" );
            error_cnt++;
            return;
          }
          //output command
          if( fwrite( p_cmd->data.cmd.data, 1, p_cmd->data.cmd.len, output_f ) != p_cmd->data.cmd.len )
          {
            printf( "ERROR: cannot write to output file\n" );
            error_cnt++;
            return;
          }
          if( flashout_f != NULL )
          {
            if( flash_out_data( p_cmd->data.cmd.data, p_cmd->data.cmd.len, flashout_f ) != p_cmd->data.cmd.len )
            {
              printf( "ERROR: cannot write to output file\n" );
              error_cnt++;
              return;
            }
          }
          pos += (unsigned __int64)p_cmd->data.cmd.len;
        }
      }
      //command was not assembled successfully
      else
      {
        printf( "INTERNAL ERROR: data of command in column %lu of line %lu is missing\n", p_cmd->create_column, p_cmd->create_line );
        error_cnt++;
      }

    } //if( p_cmd != NULL ...
  } //for( p_node ...


  //fill last sector with 0x00
  for( ; (pos & 0x1FF) != 0; pos++ )
  {
    if( fputc( 0x00, output_f ) == EOF )
      break;
    if( flashout_f != NULL )
      if( flash_out_byte( 0x00, flashout_f ) != 0 )
        break;
  }
  if( (pos & 0x1FF) != 0 )
  {
    printf( "ERROR: cannot write to output file\n" );
    error_cnt++;
    return;
  }


  //append jump to second sector
  sectors = (unsigned long)-(long)(unsigned long)(pos >> 9); //get number of sectors to jump
  i = fputc( 0x01, output_f ) == EOF; //write jump
  i |= fputc( (unsigned char)((sectors >> 24) & 0xFF), output_f ) == EOF;
  i |= fputc( (unsigned char)((sectors >> 16) & 0xFF), output_f ) == EOF;
  i |= fputc( (unsigned char)((sectors >> 8) & 0xFF), output_f ) == EOF;
  i |= fputc( (unsigned char)(sectors & 0xFF), output_f ) == EOF;
  if( flashout_f != NULL )
  {
    i |= flash_out_byte( 0x01, flashout_f ) != 0;
    i |= flash_out_byte( (unsigned char)((sectors >> 24) & 0xFF), flashout_f ) != 0;
    i |= flash_out_byte( (unsigned char)((sectors >> 16) & 0xFF), flashout_f ) != 0;
    i |= flash_out_byte( (unsigned char)((sectors >> 8) & 0xFF), flashout_f ) != 0;
    i |= flash_out_byte( (unsigned char)(sectors & 0xFF), flashout_f ) != 0;
  }
  if( i )
  {
    printf( "ERROR: cannot write to output file\n" );
    error_cnt++;
    return;
  }
  for( i = 5; i < 0x200; i++ ) //fill rest of sector with 0x00
  {
    if( fputc( 0x00, output_f ) == EOF )
      break;
    if( flashout_f != NULL )
      if( flash_out_byte( 0x00, flashout_f ) != 0 )
        break;
  }
  if( i < 0x200 )
  {
    printf( "ERROR: cannot write to output file\n" );
    error_cnt++;
    return;
  }


  //flash output: footer
  if( flashout_f != NULL )
    fwrite( "&amloaded=1", 1, 11, flashout_f );
}

//do the work
int do_it( char * input_file, char * ident_file, char * textout_file, char * flashout_file, char * output_file )
{
  FILE * input_f, * ident_f, * textout_f, * flashout_f, * output_f;


  //open input file
  input_f = fopen( input_file, "rt" );
  if( input_f == NULL )
  {
    printf( "FATAL ERROR: cannot open input file \"%s\"\n", input_file );
    return -1;
  }

  //initialize scanner
  scan_init( input_f );

  //parse input
  print_box( "parsing..." );
  yyparse( );
  print_err_warn( );
  puts( "" );
  if( p_list == NULL ) //no parse tree available
  {
    printf( "FATAL ERROR: no parse tree could be build\n" );
    puts( "" );
    return -1;
  }
  if( error_cnt > 0 ) //errors were found
    return -1;

  //close input file
  fclose( input_f );


  //check declaration of identifiers
  print_box( "checking identifiers..." );
  id_list_enum( ident_check_enum_func ); //enumerate all identifiers in identifier list
  print_err_warn( );
  puts( "" );
  if( error_cnt > 0 ) //errors were found
    return -1;


  //output identifiers
  if( ident_file != NULL )
  {
    print_box( "writing identifier file..." );

    //open identifier file
    ident_f = fopen( ident_file, "wt" );
    if( ident_f == NULL )
    {
      printf( "ERROR: cannot open identifier file \"%s\"\n", ident_file );
      error_cnt++;
    }
    else
    {
      //write identifier
      ident_write_f = ident_f;
      id_list_enum( ident_write_enum_func ); //enumerate all identifiers in identifier list

      //close identifier file
      fclose( ident_f );
    }

    print_err_warn( );
    puts( "" );
    if( error_cnt > 0 ) //errors were found
      return -1;
  }


  //assemble
  print_box( "assembling..." );
  assemble( );
  print_err_warn( );
  puts( "" );
  if( error_cnt > 0 ) //errors were found
    return -1;


  //output
  if( textout_file == NULL && flashout_file == NULL )
    print_box( "writing output file..." );
  else
    print_box( "writing output files..." );

  //open text output file
  if( textout_file == NULL )
    textout_f = NULL;
  else
  {
    textout_f = fopen( textout_file, "wt" );
    if( textout_f == NULL )
    {
      printf( "ERROR: cannot open text output file \"%s\"\n", textout_file );
      error_cnt++;
    }
  }

  //open flash output file
  if( flashout_file == NULL )
    flashout_f = NULL;
  else
  {
    flashout_f = fopen( flashout_file, "wt" );
    if( flashout_f == NULL )
    {
      printf( "ERROR: cannot open flash output file \"%s\"\n", flashout_file );
      error_cnt++;
    }
  }

  //open output file
  output_f = fopen( output_file, "wb" );
  if( output_f == NULL )
  {
    printf( "ERROR: cannot open output file \"%s\"\n", output_file );
    error_cnt++;
  }
  else
  {
    //write output
    output( textout_f, flashout_f, output_f );

    //close output file
    fclose( output_f );
  }

  //close flash output file
  if( flashout_f != NULL )
    fclose( flashout_f );

  //close text output file
  if( textout_f != NULL )
    fclose( textout_f );

  print_err_warn( );
  puts( "" );
  if( error_cnt > 0 ) //errors were found
    return -1;


  //return success
  return 0;
}

//main program
int main( int param_cnt, char * * params )
{
  int param_err, param_no, error_code;
  char * input_file, * ident_file, * textout_file, * flashout_file, * output_file;
  char * output_file_buffer;
  char * p_slash, * p_dot;

  //issue short info
  print_box2( "ARCADEmini assembler (unix)" );
  puts( "version 1.4 date 2004-03-15\n"
        "part of the ARCADEmini project - http[s]://arcademini.schuermans.info/\n"
        "Copyright (C) 2003-2004 1stein <1stein@schuermans.info>\n"
        "Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html\n"
      );

  //read parameters
  param_err = 0;
  input_file = NULL;
  ident_file = NULL;
  textout_file = NULL;
  flashout_file = NULL;
  output_file = NULL;
  for( param_no = 1; param_no < param_cnt; param_no++ )
  {
    //identifier filename
    if( strcmp( params[param_no], "-i" ) == 0 || strcmp( params[param_no], "--identifier-file" ) == 0 )
    {
      param_no++;
      if( ident_file != NULL || param_no >= param_cnt ) //already specified or no filename following
        param_err = 1;
      else
        ident_file = params[param_no];
    }
    //text output filename
    else if( strcmp( params[param_no], "-t" ) == 0 || strcmp( params[param_no], "--text-output-file" ) == 0 )
    {
      param_no++;
      if( textout_file != NULL || param_no >= param_cnt ) //already specified or no filename following
        param_err = 1;
      else
        textout_file = params[param_no];
    }
    //flash output filename
    else if( strcmp( params[param_no], "-f" ) == 0 || strcmp( params[param_no], "--flash-output-file" ) == 0 )
    {
      param_no++;
      if( flashout_file != NULL || param_no >= param_cnt ) //already specified or no filename following
        param_err = 1;
      else
        flashout_file = params[param_no];
    }
    //output filename
    else if( strcmp( params[param_no], "-o" ) == 0 || strcmp( params[param_no], "--output-file" ) == 0 )
    {
      param_no++;
      if( output_file != NULL || param_no >= param_cnt ) //already specified or no filename following
        param_err = 1;
      else
        output_file = params[param_no];
    }
    //unknown option
    else if( params[param_no][0] == '-' )
      param_err = 1;
    //input filename
    else
    {
      if( input_file != NULL ) //already specified
        param_err = 1;
      else
        input_file = params[param_no];
    }
  } //for( param_no ...

  //check parameters
  if( input_file == NULL )
    param_err = 1;

  //parameter error
  if( param_err )
  {
    //print syntax explanation
    puts( "syntax:     am_asm [options] <input-file>\n"
          "\n"
          "options:    -i / --identifier-file <identifier-file>\n"
          "            -t / --text-output-file <text-output-file>\n"
          "            -f / --flash-output-file <flash-output-file>\n"
          "            -o / --output-file <output-file>\n"
          "\n"
          "example:    am_asm -i test.amids -f test.amflash -o test.ambin test.amasm\n"
          "\n"
        );
    //exit with error
    return -1;
  }

  //no output filename
  output_file_buffer = NULL;
  if( output_file == NULL )
  {
    //allocate buffer for output-filename
    output_file_buffer =  (char *)malloc( strlen( input_file ) + 16 );
    if( output_file_buffer == NULL )
    {
      puts( "FATAL ERROR: out of memory\n" );
      return -1;
    }
    output_file = output_file_buffer; //(using extra pointer to buffer to remember to free buffer)
    //copy input filename
    strcpy( output_file, input_file );
    //try to remove extension of input filename
    p_slash = strrchr( output_file, '/' );
    if( p_slash == NULL )
      p_slash = output_file;
    p_dot = strrchr( p_slash, '.' );
    if( p_dot != NULL )
      *p_dot = 0;
    //append extension to output filename
    strcat( output_file, ".ambin" );
  }

  //do the work
  error_code = do_it( input_file, ident_file, textout_file, flashout_file, output_file );

  //print info
  if( error_code == 0 )
    print_box2( "SUCCESS" );
  else
    print_box2( "FAILED" );

  //cleanup everything
  if( output_file_buffer != NULL )
    free( output_file_buffer );
  node_free_all( );
  id_list_free_all( );

  //return error code
  return error_code;
}

